/**
  *  \file client/actions/gaexport.cc
  *  \brief Global Actions: Export
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <memory>
#include "client/actions/gaexport.h"
#include "client/dialogs/exportparam.h"
#include "cpluslib/format.h"
#include "game/error.h"
#include "game/planet.h"
#include "game/pref.h"
#include "game/ship.h"
#include "int/export.h"
#include "int/if/planetif.h"
#include "int/if/shipif.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/answer.h"
#include "util/nls.h"

/************************** WGlobalExportAction **************************/

/** Export action. The execute function just collects objects, the
    actual work is done in finish().

    The Ids of the objects to export are collected in %objects.
    The %state tracks which kind of object we're collecting, and
    detects problems. When objects have successfully been collected,
    this object also acts as an AnswerProvider used for filtering
    the export. */
class WGlobalExportAction::Executer
    : public WGlobalActionExecuter,
      private AnswerProvider
{
    enum {
        Neutral,
        CollectPlanets,
        CollectShips,
        Failure
    } state;

    enum {
        MAX_OBJS = NUM_SHIPS > NUM_PLANETS ? NUM_SHIPS : NUM_PLANETS,
        WORDS = MAX_OBJS / 32 + 1
    };

    uint32_t objects[WORDS];

 public:
    Executer()
        : state(Neutral)
        {
            std::fill(objects, objects+WORDS, uint32_t(0));
        }

    // WGlobalActionExecuter:
    bool execute(GObject& obj, bool /*lock*/);
    void finish(int obj, int success);

    // AnswerProvider:
    bool ask(int questionId, string_t question);
};

WGlobalActionExecuter*
WGlobalExportAction::prepare()
{
    return new Executer();
}

bool
WGlobalExportAction::Executer::execute(GObject& obj, bool /*lock*/)
{
    GPlanet* pl = dynamic_cast<GPlanet*>(&obj);
    GShip* sh = dynamic_cast<GShip*>(&obj);

    // Ignore anything but ships or planets.
    if (pl != 0 || sh != 0) {
        // If still neutral, infer type.
        if (state == Neutral) {
            if (pl != 0) {
                state = CollectPlanets;
            } else {
                state = CollectShips;
            }
        }

        // Make sure we only receive objects of one type.
        if (((state == CollectShips && sh != 0)
             || (state == CollectPlanets && pl != 0))
            && obj.getId() > 0
            && obj.getId() <= MAX_OBJS)
        {
            uint32_t slot = obj.getId();
            objects[slot / 32] |= 1 << (slot % 32);
        } else {
            state = Failure;
        }
    }

    return true;
}

void
WGlobalExportAction::Executer::finish(int /*obj*/, int /*success*/)
{
    // Handle failure modes
    if (state == Neutral) {
        messageBox(_("No units selected for export."), _("Export"));
        return;
    }
    if (state == Failure) {
        messageBox(_("Your selection includes ships and planets. "
                     "PCC cannot export ships and planets at the same time. "
                     "Please use the \"1\" and \"2\" buttons to select only one type."),
                   _("Export"));
        return;
    }

    // Export field list
    ConfigStringOption& opt = (state == CollectPlanets
                               ? getUserPreferences().ExportPlanetFields
                               : getUserPreferences().ExportShipFields);
    IntExportFieldList fields;
    try {
        fields.addList(opt());
    }
    catch (...)
    { }

    // Export object type
    std::auto_ptr<IntContext> type(state == CollectPlanets
                                   ? IFPlanetMake()
                                   : IFShipMake());
    if (!type.get()) {
        // This should not happen, but better be safe.
        messageBox(_("No units selected for export."), _("Export"));
        return;
    }

    // Dialog
    WExportParameterDialog dialog(fields, type.get(), *this);
    ui_root->centerWidget(dialog);
    if (ui_root->runModal(dialog)) {
        // The dialog will have performed the export. However, we still
        // have to remember the field list for next time.
        opt.set(fields.toString());

        // Information. We only get here when export really succeeded;
        // exceptions are caught by onOK and keep the dialog open.
        messageBox(_("Export succeeded."), _("Export"));
    }
}

bool
WGlobalExportAction::Executer::ask(int questionId, string_t /*question*/)
{
    return questionId > 0
        && questionId <= MAX_OBJS
        && (objects[questionId / 32] & (1 << (questionId % 32))) != 0;
}
