/**
  *  \file client/dialogs/baseammo.cc
  *  \brief Starbase Build Ammunition Dialog
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/baseammo.h"
#include "client/wcost.h"
#include "client/widgets/ammoinfo.h"
#include "client/widgets/ammolist.h"
#include "client/widgets/ftrinfo.h"
#include "client/winfo.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "cpluslib/string.h"
#include "game/actions/baseammo.h"
#include "game/actions/basetech.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/reg.h"
#include "game/ship.h"
#include "game/spec.h"
#include "game/team.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/prefix.h"
#include "ui/widgets/button.h"
#include "ui/widgets/cardgroup.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"
#include "util/misc.h"
#include "ui/widgets/quit.h"

namespace {
    class WAmmoWindow : public UIWindow {
        ObjectHolder              h;
        GMultiTransaction         mt;
        GCargoContainer*          receiver;
        GStarbaseBuildAmmoAction* action;
        WTorpInfo*                torp_info;
        WFighterInfo*             ftr_info;
        WAmmoList*                list;
        WAmmoInfo*                info;
        WObservingItemCostDisplay* dpy;

        void init();
        void build(int32 amount);
     public:
        WAmmoWindow(GPlanet& p);
        WAmmoWindow(GPlanet& pay, GShip& receive);

        bool handleEvent(const UIEvent& e, bool second_pass);
        void scrolled();
        void enter();
        void cancel();
    };
};

/****************************** WAmmoWindow ******************************/

WAmmoWindow::WAmmoWindow(GPlanet& p)
    : UIWindow(_("Build Torpedoes/Fighters"), BLUE_WINDOW, COLOR_GRAY)
{
    GPlanetTransfer& pt = mt.addNewSubtransaction(new GPlanetTransfer(p));
    this->receiver = &pt;
    this->action = &mt.addNewSubtransaction(new GStarbaseBuildAmmoAction(pt, pt));
    init();
}

WAmmoWindow::WAmmoWindow(GPlanet& pay, GShip& receive)
    : UIWindow(_("Build Torpedoes/Fighters"), BLUE_WINDOW, COLOR_GRAY)
{
    GPlanetTransfer& pt = mt.addNewSubtransaction(new GPlanetTransfer(pay));
    GShipTransfer& st = mt.addNewSubtransaction(new GShipTransfer(receive));
    this->receiver = &st;
    this->action = &mt.addNewSubtransaction(new GStarbaseBuildAmmoAction(pt, st));
    init();
}

void
WAmmoWindow::init()
{
    // VBox
    //   HBox
    //     VBox
    //       "Weapon:"
    //       UIFrameGroup
    //         WAmmoList list
    //     VBox
    //       WAmmoInfo info
    //       UICardGroup
    //         WTorpInfo
    //         WFighterInfo
    //   "Total Cost:"
    //   HBox
    //     WObservingItemCostDisplay dpy
    //   HBox
    //     "OK"
    //     "Cancel"
    //     UISpacer
    //     "Help"
    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g11 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g12 = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g2  = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g3  = h.add(new UIGroup(UIHBoxLayout::instance5));

    list = &h.add(new WAmmoList(*action));
    info = &h.add(new WAmmoInfo(*receiver, action->getCargoTypeFromIndex(list->getCurrent()+1)));

    g11.add(h.add(new UIStaticText(_("Weapon:"), UIStaticText::StaticTitle)));
    g11.add(UIFrameGroup::wrap(h, *list));

    ftr_info  = &h.add(new WFighterInfo());
    torp_info = &h.add(new WTorpInfo());
    UICardGroup& g121 = h.add(new UICardGroup());
    g121.add(*ftr_info);
    g121.add(*torp_info);
    g12.add(*info);
    g12.add(g121);

    g1.add(g11);
    g1.add(g12);

    dpy = &h.add(new WObservingItemCostDisplay(*action, action->getPlanet(), true));
    g2.add(*dpy);

    UIButton& ok = h.add(new UIButton(1000,  _("OK"), SDLK_RETURN));
    UIButton& esc = h.add(new UIButton(1001, _("Cancel"), SDLK_ESCAPE));
    g3.add(ok);
    g3.add(esc);
    g3.add(h.add(new UISpacer()));
    add(g1);
    add(h.add(new UIStaticText(_("Total Cost:"), UIStaticText::StaticTitle)));
    add(g2);
    add(g3);
    add(h.add(new UIPrefixArg()));
    add(h.add(new UIQuit(1001)));

    ok.sig_fired.add(this, &WAmmoWindow::enter);
    esc.sig_fired.add(this, &WAmmoWindow::cancel);
    list->sig_moved.add(this, &WAmmoWindow::scrolled);
    list->scrollTo(NUM_TORPS+1, -1);

    pack();
}

void
WAmmoWindow::build(int32 amount)
{
    int index = list->getCurrent()+1;

    // first, if we're trying to build, and don't have sufficient
    // tech, we have to upgrade.
    if (amount == 0)
        return;
    if (amount > 0) {
        if (int tech = action->getTechUpgradeNeeded(index)) {
            // can we upgrade tech?
            // FIXME: this should probably be converted to use GMultiTransaction
            GPlanetTransfer tech_pay(action->getPlanet());
            GStarbaseTechUpgradeAction tech_action(action->getPlanet());
            GStarbaseBuildTransaction ta(tech_action, tech_pay);
            if (!tech_action.setTech(tl_Torpedo, tech)) {
                messageBox(_("Building this torpedo requires you to upgrade your tech level over Tech 6. "
                             "As an unregistered player, you can't do that."),
                           _("Tech Upgrade"));
                ta.doCancel();
                tech_pay.doCancel();
                return;
            }
            if (ta.getError()) {
                messageBox(format(_("Building this torpedo requires a tech upgrade. "
                                    "You do not have the %d megacredits required for the upgrade."),
                                  ta.getCost(el_Money)),
                           _("Tech Upgrade"));
                ta.doCancel();
                tech_pay.doCancel();
                return;
            }
            if (UIMessageBox(format(_("Building this torpedo requires tech %d. "
                                      "Do you want to upgrade for %d megacredits?"),
                                    tech,
                                    numToString(ta.getCost(el_Money))),
                             _("Tech Upgrade")).doYesNoDialog())
            {
                ta.doCommit();
                tech_pay.doCommit();
            } else {
                ta.doCancel();
                tech_pay.doCancel();
            }

            // do not immediately buy stuff after upgrading tech
            // FIXME: widgets should figure out itself that it needs to
            // be redrawn
            dpy->drawWidget(false);
            list->drawWidget(false);
            scrolled();         // this will update the tech level
            getDisplayedTurn().getCurrentUniverse().doScreenUpdates();
            return;
        }
    }

    // okay, we're building stuff
    action->setAmount(index, action->getAmount(index) + action->limitAmount(index, amount));
    dpy->drawWidget(false);
    list->drawWidget(false);
}

bool
WAmmoWindow::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case '+':
            if (int i = e.key.arg)
                build(i);
            else
                build(1);
            return true;
         case '-':
            if (int i = e.key.arg)
                build(-i);
            else
                build(-1);
            return true;
         case ss_Ctrl + '+':
            build(100);
            return true;
         case ss_Ctrl + '-':
            build(-100);
            return true;
         case ss_Alt + '+':
            build(10000);
            return true;
         case ss_Alt + '-':
            build(-10000);
            return true;
        }
    }
    return UIWindow::handleEvent(e, second_pass);
}

void
WAmmoWindow::scrolled()
{
    const int what = list->getCurrent() + 1;

    info->setCargoType(action->getCargoTypeFromIndex(what));

    // Show spec item
    int tech = action->getPlanet().getBaseTechLevel(tl_Torpedo);
    if (what > 0 && what <= NUM_TORPS) {
        // Torpedo
        const GTorpedo& torp = getTorp(what).getTorpedo();
        dpy->setPartCost(torp.getCost());
        dpy->setPartTech(tech, torp.getTech());
        torp_info->setTorpId(what);
        torp_info->getParent()->setFocusedChild(*torp_info);
    } else {
        // Fighter
        int owner = action->getPlanet().getOwner();
        GFighter f = getFighterForPlayer(owner);
        dpy->setPartCost(f.getCost());
        dpy->setPartTech(tech, f.getTech());
        ftr_info->setPlayerId(owner);
        ftr_info->getParent()->setFocusedChild(*ftr_info);
    }
}

void
WAmmoWindow::enter()
{
    if (!mt.getError()) {
        mt.doCommit();
        stop(1);
    }
}

void
WAmmoWindow::cancel()
{
    mt.doCancel();
    stop(0);
}

void
doBaseBuildAmmo(GPlanet& p)
{
    WAmmoWindow aw(p);
    ui_root->centerWidget(aw);
    ui_root->runModal(aw);
}

void
doBaseBuildAmmo(GPlanet& p, GShip& s)
{
    WAmmoWindow aw(p, s);
    ui_root->centerWidget(aw);
    ui_root->runModal(aw);
}
