/**
  *  \file client/wscore.h
  *  \brief Score Display
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_CLIENT_WSCORE_H
#define PCC_V2_CLIENT_WSCORE_H

#include <vector>
#include "cpluslib/string.h"
#include "game/stat.h"
#include "game/struct.h"
#include "ui/ui.h"
#include "game/playerset.h"
#include "ui/widgets/tooltip.h"

enum WScoreAggregation {
    ScoreByPlayer,
    ScoreByTeam,
    ScoreByPlayerCumulative,
    ScoreByTeamCumulative
};

inline WScoreAggregation makeCumulative(WScoreAggregation agg)    { return WScoreAggregation(agg | 2); }
inline WScoreAggregation withoutCumulative(WScoreAggregation agg) { return WScoreAggregation(agg & 1); }
inline bool              isCumulative(WScoreAggregation agg)      { return (agg & 2) != 0;             }
inline WScoreAggregation toggleTeam(WScoreAggregation agg)        { return WScoreAggregation(agg ^ 1); }
inline WScoreAggregation toggleCumulative(WScoreAggregation agg)  { return WScoreAggregation(agg ^ 2); }

/** Definition of a score column. Allows a linear-combination of several columns
    to be set up. This allows display of common scores like Tim's score or
    "total ships" from a regular score recording. */
class WScore {
    enum { NUM = 4 };
    string_t   name;
    int        num_parts;
    GStatIndex columns[NUM];
    int        factors[NUM];

 public:
    WScore(string_t name)
        : name(name), num_parts(0) { }
    WScore& add(GStatIndex column, int factor);
    int32 get(const GStatRecord& record, int player, WScoreAggregation aggr) const;
    int32 getTotal(const GStatRecord& record) const;
    int32 getMax(const GStatFile& file, uint32 first, uint32 last, WScoreAggregation aggr) const;
    GPlayerSet getPresence(const GStatFile& file, uint32 first, uint32 last, WScoreAggregation aggr) const;
    const string_t& getName() const
        { return name; }
    int getNumParts() const
        { return num_parts; }
    GStatIndex getPart(int i) const
        { return columns[i]; }
    int getFactor(int i) const
        { return factors[i]; }
};

/** Score table display. Displays a configurable set of columns of a selectable
    turn from a GStatFile. Can display differences and percentages, and sort
    display by each column. */
class WScoreTable : public UIWidget {
 public:
    enum Mode {
        Normal,                     /**< Display normal scoreboard. */
        DiffToPrevious,             /**< Display differences to respective predecessor. */
        DiffToParticular,           /**< Display differences to specific turn. */
        Percentages,                /**< Display percentages of selected. */
        RatioOfTotal                /**< Display percentages of total. */
    };
    enum { NUM_MODES = 5 };

    WScoreTable(const GStatFile& file);

    void addColumn(const WScore& score);

    void addDefaultColumns();

    int  getNumColumns() const;

    void setSortOrder(int order);

    void setMode(Mode mode);

    Mode getMode() const;

    void setAggregation(WScoreAggregation aggr);

    WScoreAggregation getAggregation() const;

    void setCurrentPlayer(int player);

    int  getCurrentPlayer() const;

    void setCurrentRecord(uint32 record);

    uint32 getCurrentRecord() const;

    uint32 getComparedRecord() const;

    void getLayoutInfo(LayoutInfo& info);

    void drawContent(GfxCanvas& can);

    bool handleEvent(const UIEvent& evt, bool second_pass);

    void onStateChange(int bits, bool enabled);

    string_t getModeName(Mode mode, uint32 compare);

    // User actions:
    void nextTurn();
    void previousTurn();
    void nextPlayer();
    void previousPlayer();

    Signal0 sig_turn_changed;
    Signal0 sig_config_changed;

 private:
    const GStatFile& file;          /**< Data source. */

    std::vector<WScore> columns;    /**< Definition of the columns. */
    int players[NUM_PLAYERS];       /**< Current sort order. */
    int current_player;             /**< Current player (NOT line!). */
    Mode mode;                      /**< Mode of display. */
    WScoreAggregation aggreg;       /**< Mode of aggregation. */
    int sort_order;                 /**< Sort order. Index into columns to sort by that column; -1 to sort by player only. */

    int label_width;                /**< Width of player-name column. */
    int column_width;               /**< Width of score column. */
    int header_height;              /**< Height of header line. */
    int line_height;                /**< Height of data line. */
    int footer_height;              /**< Height of footer. */

    bool mouse_down;                /**< True iff mouse button is currently down. */

    uint32 current_record;          /**< Currently-displayed record. */
    uint32 compare_record;          /**< Record to compare to. */

    void updateSortOrder();
};

/** Score table config display. Displays the settings chosen in the score
    table display. */
class WScoreTableConfigDisplay : public UIWidget {
    WScoreTable& table;
 public:
    WScoreTableConfigDisplay(WScoreTable& table);

    void getLayoutInfo(LayoutInfo& info);

    void drawContent(GfxCanvas& can);
};

class WScoreTableTimeDisplay : public UIWidget {
    WScoreTable& table;
    const GStatFile& file;          /**< Data source. */
 public:
    WScoreTableTimeDisplay(WScoreTable& table, const GStatFile& file);

    void getLayoutInfo(LayoutInfo& info);

    void drawContent(GfxCanvas& can);
};

/** Score chart. Displays a configurable score over a configurable range of
    records from a GStatFile. One player's curve can be highlighted using a
    bold curve. */
class WScoreChart : public UIWidget {
 public:
    WScoreChart(const GStatFile& file, uint32 first, uint32 last);

    void setScore(WScore score);

    void setAggregation(WScoreAggregation aggr);

    WScoreAggregation getAggregation() const;

    void setCurrentPlayer(int player);

    int  getCurrentPlayer() const;

    void getLayoutInfo(LayoutInfo& info);

    void drawContent(GfxCanvas& can);

    bool handleEvent(const UIEvent& e, bool second_pass);

    void onStateChange(int state, bool enable);

 private:
    const GStatFile& file;          /**< Data source. */
    uint32 first;                   /**< First record to display. */
    uint32 last;                    /**< One-past-last record to display. */
    WScore score;                   /**< Score to display. */
    int current_player;             /**< Current player. */
    WScoreAggregation aggreg;       /**< Mode of aggregation. */

    int left_margin;                /**< Left margin. Contains "Turn 1" and "nnnn" (max) display. */
    int right_margin;               /**< Right margin. Contains "Turn nn" display. */
    int bottom_margin;              /**< Bottom margin. Contains second line of "Turn nn" display. */

    int32 max_score;                /**< Maximum score on grid. */

    int32 victory_level;              /**< Victory level to draw, if any. Negative if none. */
    int16 victory_turns_needed;       /**< Number of turns required for victory. */
    GPlayerArray<int> victory_turns;  /**< Turn when victory limit reached, for each player. Real turn numbers, not indexes. 0 if never. */

    int aux_decay;                  /**< Decay rate of auxiliary grid, percentage per turn. Zero if regular grid without decay. */
    int32 aux_scale;                /**< Distance of auxiliary grid lines. */
    int32 aux_limit;                /**< Limit of auxiliary grid lines. */

    GPlayerSet present_curves;      /**< List of present players or teams. */

    UITooltipControl tt_ctl;        /**< Tooltip control. */

    void recomputeGrid();

    void drawPlayerChart(GfxContext& ctx, int player);
    void drawPlayerShade(GfxContext& ctx, int player);

    void findVictoryCondition();

    void doTooltip(GfxPoint pt);
};

inline int
WScoreTable::getCurrentPlayer() const
{
    return current_player;
}

inline WScoreTable::Mode
WScoreTable::getMode() const
{
    return mode;
}

inline WScoreAggregation
WScoreTable::getAggregation() const
{
    return aggreg;
}

inline uint32
WScoreTable::getCurrentRecord() const
{
    return current_record;
}

inline uint32
WScoreTable::getComparedRecord() const
{
    return compare_record;
}

inline WScoreAggregation
WScoreChart::getAggregation() const
{
    return aggreg;
}

inline int
WScoreChart::getCurrentPlayer() const
{
    return current_player;
}

#endif
