/**
  *  \file flak/flak-vis.cc
  *  \brief FLAK Visualisation Stubs
  *
  *  Visualisation base classes.
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "flak/flak.h"
#include "cpluslib/assert.h"

/*! \class FlakVisualizer
    A FlakVisualizer descendant is used to display a FLAK battle. No battle
    can be played without a visualizer. FlakBattle will call FlakVisualizer's
    methods at appropriate places.

    The calling sequences are documented for each member function. You
    might need to carry some status, but we hope this is at a
    minimum. */

/** Constructor. This registers the visualizer with the battle. Note
    that, at any given time, only one visualizer might be registered
    for a battle. */
FlakVisualizer::FlakVisualizer(FlakBattle& the_battle)
    : the_battle(the_battle)
{
    ASSERT(!the_battle.getVisualizer());
    the_battle.setVisualizer(this);
}

/** Destructor. Unregisters the visualizer from the battle. */
FlakVisualizer::~FlakVisualizer()
{
    if (the_battle.getVisualizer() == this)
        the_battle.setVisualizer(0);
}

/*************************** FlakNullVisualizer **************************/

/*! \class FlakNullVisualizer
    This visualizer ignores all requests, and can therefore be used for
    silent simulations. */

void
FlakNullVisualizer::createObject(FlakObject& /*obj*/, ship_t /*source*/)
{ }

void
FlakNullVisualizer::destroyObject(FlakObject& /*obj*/, bool /*violent*/)
{ }

void
FlakNullVisualizer::destroyShip(ship_t /*unit*/)
{ }

void
FlakNullVisualizer::fireBeam(FlakObject& /*obj1*/, FlakObject& /*obj2*/)
{ }

void
FlakNullVisualizer::fireBeam(FlakObject& /*obj1*/, ship_t /*unit2*/)
{ }

void
FlakNullVisualizer::fireBeam(ship_t /*unit1*/, FlakObject& /*obj2*/, bool /*hits*/)
{ }

void
FlakNullVisualizer::fireBeam(ship_t /*unit1*/, ship_t /*unit2*/, bool /*hits*/)
{ }

void
FlakNullVisualizer::hitShip(ship_t /*unit*/, int /*expl*/, int /*kill*/, int /*death_flag*/)
{ }
