/**
  *  \file game/actions/cargocost.cc
  *  \brief Cargo Cost Transaction
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/actions/cargocost.h"
#include "game/cargo.h"
#include "util/nls.h"

/************************* GCargoCostTransaction *************************/

/** Constructor. Initialize with cost zero.
    \param container the container that pays the transaction. It should
    be a planet; we assume that addCargo never fails. */
GCargoCostTransaction::GCargoCostTransaction(GCargoContainer& container)
    : container(container)
{ }

/** Destructor. Cancels the transaction if not already finished. */
GCargoCostTransaction::~GCargoCostTransaction()
{
    if (!isFinished())
        doCancel();
}

/** Check if transaction is valid. The transaction can only be
    committed when this returns true.
    \inv isValid() <=> getMissing(T) == 0 for all valid Ts */
bool
GCargoCostTransaction::isValidCost() const
{
    return reserved.isEnoughFor(cost);
}

/** Update our reservations. */
void
GCargoCostTransaction::update()
{
    // Minerals are simple:
    reserved.add(el_Tritanium,  -container.changeCargo(el_Tritanium,  reserved.get(el_Tritanium)  - cost.get(el_Tritanium),  true));
    reserved.add(el_Duranium,   -container.changeCargo(el_Duranium,   reserved.get(el_Duranium)   - cost.get(el_Duranium),   true));
    reserved.add(el_Molybdenum, -container.changeCargo(el_Molybdenum, reserved.get(el_Molybdenum) - cost.get(el_Molybdenum), true));

    // Cash and supplies are a little harder. So we simply give back everything
    // and re-acquire it:
    container.changeCargo(el_Supplies, reserved.get(el_Supplies), false);
    container.changeCargo(el_Money,    reserved.get(el_Money),    false);
    reserved.set(el_Money, -container.changeCargo(el_Money, -cost.get(el_Money), true));
    reserved.set(el_Supplies,
                 -container.changeCargo(el_Supplies,
                                        -(cost.get(el_Supplies)
                                          + (cost.get(el_Money) - reserved.get(el_Money))),
                                        true));
}

/** Change the cost.
    \param type type of resource to modify
    \param amount amount to add (negative to subtract)
    \see setCost() */
void
GCargoCostTransaction::changeCost(GCargoType type, int32_t amount)
{
    ASSERT(GCost::isValidType(type));
    cost.add(type, amount);
    update();
    sig_cost_changed.raise();
}

/** Change the cost.
    \param cost cost to add/remove
    \see setCost() */
void
GCargoCostTransaction::changeCost(const GCost& cost)
{
    this->cost += cost;
    update();
    sig_cost_changed.raise();
}

/** Set cost. This does actual re-allocation.
    \post getCost(type) == amount */
void
GCargoCostTransaction::setCost(GCargoType type, int32_t amount)
{
    ASSERT(GCost::isValidType(type));
    cost.set(type, amount);
    update();
    sig_cost_changed.raise();
}

/** Set cost. Takes a GCost instead of individual components. */
void
GCargoCostTransaction::setCost(const GCost& cost)
{
    this->cost = cost;
    update();
    sig_cost_changed.raise();
}

/** Get cost. */
int32_t
GCargoCostTransaction::getCost(GCargoType type) const
{
    return cost.get(type);
}

/** Get cost. */
GCost
GCargoCostTransaction::getCost() const
{
    return cost;
}

/** Get remaining cargo after this transaction. */
int32_t
GCargoCostTransaction::getRemainingCargo(GCargoType type) const
{
    /* The container's getCargo will have all reservations,
       including ours, subtracted */
    ASSERT(GCost::isValidType(type));
    return container.getCargo(type);
}

/** Get missing stuff for this transaction. When there's something
    missing, isValid() will be false.
    \inv isValid() => return value will be 0 */
int32_t
GCargoCostTransaction::getMissingCargo(GCargoType type) const
{
    ASSERT(GCost::isValidType(type));
    if (type == el_Supplies) {
        /* missing supplies. We're lacking supplies only when the cost
           actually includes some. */
        if (cost.get(el_Supplies) <= reserved.get(el_Supplies))
            return 0;
        else
            return cost.get(el_Supplies) - reserved.get(el_Supplies);
    } else if (type == el_Money) {
        /* missing money. When we have too little money but can
           compensate with supplies, pretend we're not lacking cash. */
        int32 unused_supplies = reserved.get(el_Supplies) - cost.get(el_Supplies);
        if (unused_supplies < 0)
            unused_supplies = 0;
        int32 missing_money = cost.get(el_Money) - reserved.get(el_Money) - unused_supplies;
        if (missing_money < 0)
            return 0;           // FIXME: can this happen?
        else
            return missing_money;
    } else {
        /* minerals don't convert */
        return cost.get(type) - reserved.get(type);
    }
}

/** Get reserved amounts of a given type.
    If the transaction is successful, this value is equal to the
    actual cost of the transaction, but in case of supplies sold for
    money, it yields the actual amounts used. */
int32_t
GCargoCostTransaction::getReservedCargo(GCargoType type) const
{
    ASSERT(GCost::isValidType(type));
    return reserved.get(type);
}

/** Commit transaction. Caller must commit the GCargoContainer to actually
    use the minerals (and do the thing this cost is for, of course). */
void
GCargoCostTransaction::commit()
{
    container.mustCommit();
}

/** Cancel transaction. Caller must cancel the GCargoContainer, too. */
void
GCargoCostTransaction::cancel()
{
    container.mustCancel();
}

/** Get error message. */
GError
GCargoCostTransaction::getError()
{
    if (!isValidCost())
        return GError(GError::eNoResource, _("Not enough resources to perform this action"));
    return GError();
}
