/**
  *  \file game/fcode-list.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include <cctype>
#include "game/fcode.h"
#include "cpluslib/string.h"

/** Compare friendly codes. Alpha-numerical codes sort before those
    with other characters, and codes are sorted case-blind. */
static bool
compareFCodes(const GFCode& l, const GFCode& r)
{
    using namespace std;
    const string_t ls = l.getFCode(), rs = r.getFCode();

    if (ls.length() && rs.length()) {
        if (isalnum(ls[0]) && !isalnum(rs[0]))
            return true;
        else if (!isalnum(ls[0]) && isalnum(rs[0]))
            return false;
    }

    int i = strCaseCmp(ls, rs);
    if (i == 0)
        return ls < rs;
    else
        return i < 0;
}

/******************************* GFCodeList ******************************/

/** Static instance (the fcodes.cc file). */
GFCodeList GFCodeList::instance;

/** Make empty list. */
GFCodeList::GFCodeList()
{ }

/** Destructor. */
GFCodeList::~GFCodeList()
{ }

/** Add a friendly code. You should use sort as soon as possible to
    re-sort the list. */
void
GFCodeList::addFCode(const GFCode& code)
{
    data.push_back(code);
}

/** Sort friendly-code list in a sensible way to present it to
    users. */
void
GFCodeList::sort()
{
    /* FIXME: use stable_sort instead? */
    std::sort(data.begin(), data.end(), compareFCodes);
}

/** Clear this list. */
void
GFCodeList::clear()
{
    data.clear();
}

/** Look up friendly code by name.
    \returns iterator pointing to friendly code, or end() if none */
GFCodeList::iterator
GFCodeList::getFCodeByName(const string_t fc) const
{
    iterator i = begin();
    while (i != end() && i->getFCode() != fc)
        ++i;
    return i;
}

/** True iff friendly code is special. A friendly code is special if
    it is on the global FC list and not marked unspecial.
    \param fc          the FCode to test
    \param ignore_case true to ignore case. HOST has some codes
                       case-insensitive, so we need case-insensitive
                       comparison to avoid triggering a special action
                       when generating random FCs */
bool
GFCode::isSpecial(const string_t& fc, bool ignore_case)
{
    GFCodeList::iterator i = GFCodeList::getInstance().begin(),
        end = GFCodeList::getInstance().end();
    while (i != end) {
        if ((ignore_case ? strCaseCmp(fc, i->getFCode()) == 0 : fc == i->getFCode())
            && (i->getFlags() & fc_Unspecial) == 0)
        {
            return true;
        }
        ++i;
    }
    return false;
}
