/**
  *  \file game/sim.h
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */
#ifndef PCC_V2_GAME_SIM_H
#define PCC_V2_GAME_SIM_H

#include <vector>
#include "cpluslib/object.h"
#include "cpluslib/sigslot.h"
#include "game/spec.h"
#include "game/playerset.h"
#include "game/vcr.h"

class Stream;

/*! \defgroup Simulator Combat Simulator
    Classes to manage data for the combat simulator, run the simulations,
    and evaluate them. */
/*@{*/

class GSimShip;
class GSimPlanet;
class GSimState;

/// Simulator hull function.
enum GSimFunction {
    sf_PlanetImmunity,
    sf_FullWeaponry,
    sf_Commander,
    sf_TripleBeamKill,
    sf_DoubleBeamCharge,
    sf_DoubleTorpCharge,
    sf_Elusive,
    sf_Squadron
};

class GSimObject : public Object {
    // common
    int  id;
    char name[20];
    char fcode[3];
    int  damage;
    int  shield;
    int  owner;
    int  experience_level;
    int32_t flags;
    int32_t flak_r_override;
    int  flak_c_override;

    bool changed;

    VcrStatItem stats;

 protected:
    void setChanged();

    friend class GSimState;
    void setUnchanged();
    bool isChanged() const;

 public:
    GSimObject();
    ~GSimObject();

    int  getId() const;
    void setId(int id);

    string_t getName() const;
    void     setName(string_t name);

    string_t getFCode() const;
    void     setFCode(string_t fcode);

    int  getDamage() const;
    void setDamage(int damage);

    int  getShield() const;
    void setShield(int shield);

    int  getOwner() const;
    void setOwner(int owner);

    int  getExperienceLevel() const;
    void setExperienceLevel(int experience_level);

    int32_t getFlags() const;
    void setFlags(int32_t flags);

    int32 getFlakRatingOverride() const;
    void  setFlakRatingOverride(int32 r);

    int32 getFlakCompensationOverride() const;
    void  setFlakCompensationOverride(int32 r);

    bool hasFunction(GSimFunction simf) const;
    virtual bool hasImpliedFunction(GSimFunction simf) const = 0;
    bool hasAnyNonstandardFunction() const;

    VcrStatItem& getStat()
        { return stats; }
    const VcrStatItem& getStat() const
        { return stats; }

    // derived information
    int getMaxDamage() const;
    int getMaxShield() const;

    void assignRandomFCode();
    bool assignRandomFCodeFlags();

    /** Flag bits. */
    enum {
        fl_RandomFC          = 1,    ///< Friendly code randomisation enabled.
        fl_RandomFC1         = 2,    ///< Randomize first place of fcode.
        fl_RandomFC2         = 4,    ///< Randomize second place of fcode.
        fl_RandomFC3         = 8,    ///< Randomize third place of fcode.
        fl_RatingOverride    = 16,   ///< Use FLAK rating overrides.
        fl_Cloaked           = 32,   ///< Ship is cloaked.
        fl_Deactivated       = 64,   ///< Ignore this ship during simulation.
        fl_PlanetImmunity    = 128,  ///< Ship has Planet Immunity.
        fl_PlanetImmunitySet = 256,  ///< PlanetImmunity bit is effective.
        fl_FullWeaponry      = 512,  ///< Ship has Full Weaponry.
        fl_FullWeaponrySet   = 1024, ///< FullWeaponry bit is effective.
        fl_Commander         = 2048, ///< Ship is Commander.
        fl_CommanderSet      = 4096, ///< Commander bit is effective.

        fl_TripleBeamKill      = 1*65536,     ///< Ship has 3x beam kill.
        fl_TripleBeamKillSet   = 2*65536,     ///< TripleBeamKill bit is effective.
        fl_DoubleBeamCharge    = 4*65536,     ///< Ship has 2x beam recharge.
        fl_DoubleBeamChargeSet = 8*65536,     ///< DoubleBeamCharge bit is effective.
        fl_DoubleTorpCharge    = 16*65536,    ///< Ship has 2x torp recharge.
        fl_DoubleTorpChargeSet = 32*65536,    ///< DoubleTorpCharge bit is effective.
        fl_Elusive             = 64*65536,    ///< Ship sets enemy's torp hit odds to 10%
        fl_ElusiveSet          = 128*65536,   ///< Elusive bit is effective.
        fl_Squadron            = 256*65536,   ///< Ship is a fighter squadron (unkillable, respawns).
        fl_SquadronSet         = 512*65536,   ///< Squadron bit is effective.

        fl_RandomDigits      = fl_RandomFC1 + fl_RandomFC2 + fl_RandomFC3,
        fl_FunctionSetBits   = (fl_PlanetImmunitySet | fl_CommanderSet | fl_FullWeaponrySet
                                | fl_TripleBeamKillSet | fl_DoubleBeamChargeSet | fl_DoubleTorpChargeSet
                                | fl_ElusiveSet | fl_SquadronSet)
    };

    static const uint32_t simf_to_value_bit[], simf_to_set_bit[];
};

class GSimShip : public GSimObject {
    // ships
    int  crew;
    int  hull;
    int  mass;                  /* custom ships */
    int  beam_type, beam_count;
    int  torp_type, torp_lcount;
    int  bay_count;
    int  ammo;
    int  engine_type;
    int  aggressiveness;
    int  intercept_id;

 public:
    GSimShip();
    GSimShip(const TSimShipData& file_data, int version);
    ~GSimShip();

    void storeData(TSimShipData& file_data, int version);

    int  getCrew() const;
    void setCrew(int crew);

    int  getHull() const;
    void setHull(int hull);

    int  getMass() const;
    void setMass(int mass);

    int  getBeamType() const;
    void setBeamType(int beam_type);
    int  getBeamCount() const;
    void setBeamCount(int beam_count);

    int  getTorpType() const;
    void setTorpType(int torp_type);
    int  getTorpLauncherCount() const;
    void setTorpLauncherCount(int torp_lcount);

    int  getBayCount() const;
    void setBayCount(int bay_count);

    int  getAmmo() const;
    void setAmmo(int ammo);

    int  getEngineType() const;
    void setEngineType(int engine_type);
    
    int  getAggressiveness() const;
    void setAggressiveness(int aggressiveness);

    int  getInterceptId() const;
    void setInterceptId(int id);

    // derived
    bool isCustomShip() const;

    bool isMatchingShipList() const;

    // Hull functions
    virtual bool hasImpliedFunction(GSimFunction simf) const;

    /** Aggressiveness values. Values between 1 and 12 correspond to primary
        enemy settings. */
    enum {
        /// "Kill" mission.
        agg_Kill    = -1,
        /// Passive unit.
        agg_Passive = 0,
        /// Unit has no fuel.
        agg_NoFuel  = 13
    };
};

class GSimPlanet : public GSimObject {
    // planets
    int  defense;
    int  base_defense;
    int  beam_tech;             /* 0 = no base */
    int  torp_tech;
    int  base_fighters;
    int  base_torps[NUM_TORPS];

 public:
    GSimPlanet();
    GSimPlanet(const TSimPlanetData& file_data, int version);
    ~GSimPlanet();

    void storeData(TSimPlanetData& file_data, int version);

    int  getDefense() const;
    void setDefense(int defense);

    int  getBaseDefense() const;
    void setBaseDefense(int base_defense);

    int  getBaseBeamTech() const;
    void setBaseBeamTech(int beam_tech);

    int  getBaseTorpTech() const;
    void setBaseTorpTech(int torp_tech);

    int  getBaseFighters() const;
    void setBaseFighters(int base_fighters);

    int  getBaseTorps(int tech) const;
    void setBaseTorps(int tech, int amount);

    int  getBaseTorpsAsType(int tech) const;

    bool hasBase() const
        { return getBaseBeamTech() != 0; }

    // FIXME!!!!!!!!!!!!!!!!!1111
    int  getBaseDamage() const
        { return 0; }
    void setBaseDamage(int)
        { }

    virtual bool hasImpliedFunction(GSimFunction simf) const;
};

struct GSimOptions {
    /** Balancing mode. Various algorithms to correct the TVCR
        left/right imbalance. */
    enum BalancingMode {
        BalanceNone,            ///< No balancing.
        Balance360k,            ///< 360kt bonus (HOST).
        BalanceMasterAtArms     ///< Master at Arms proposal.
    };
    /** VCR mode. Which VCR is used to run the simulation. */
    enum VcrMode {
        VcrHost,                ///< HOST.
        VcrPHost2,              ///< PHost 2.x.
        VcrPHost3,              ///< PHost 3.x.
        VcrPHost4,              ///< PHost 4.x.
        VcrFLAK,                ///< FLAK.
        VcrNuHost               ///< NuHost.
    };

    enum {
        BalanceMAX = BalanceMasterAtArms,
        VcrMAX = VcrNuHost
    };

    GPlayerBitMatrix alliance_settings;
    GPlayerBitMatrix enemy_settings;

    int            engine_shield_bonus;
    bool           scotty_bonus;
    bool           random_left_right;
    bool           honor_alliances;
    bool           only_one_simulation;
    bool           seed_control;
    bool           randomize_fcodes_on_every_fight;
    BalancingMode  balancing_mode;
    VcrMode        vcr_mode;

    int            seed;

    GSimOptions();

    bool isExperienceEnabled() const;

    void loadDefaults(VcrMode mode);
    void setMode(VcrMode mode);
};

class GSimState {
    std::vector<GSimShip> ships;
    GSimPlanet            planet;
    bool                  has_planet;
    bool                  structure_change;
 public:
    typedef std::vector<GSimShip>::size_type ship_index_t;
    
    GSimState();
    GSimState(const GSimState& other);
    ~GSimState();

    GSimState& operator=(const GSimState& other);

    bool hasPlanet() const
        { return has_planet; }
    void addPlanet();
    void removePlanet();

    ship_index_t addShip();
    ship_index_t getNumShips() const;
    ship_index_t getIndexOf(const GSimShip& sh) const;
    ship_index_t getIndexOf(int id) const;
    void duplicateShip(ship_index_t ix, int id);
    void removeShip(ship_index_t ix);
    void swapShips(ship_index_t a, ship_index_t b);
    bool isShipIdInUse(int sid) const;
    const GSimShip* getShipById(int id) const;
    GSimShip* getShipById(int id);
    int getFreeId();

    bool isMatchingShipList() const;

    void initializeStats();

    void doUpdates();

    void load(Stream& in);
    void save(Stream& out);

    void swap(GSimState& other);

    void assignRandomFCodes();

    GSimPlanet& getPlanet();
    const GSimPlanet& getPlanet() const;

    GSimShip& getShip(ship_index_t sid);
    const GSimShip& getShip(ship_index_t sid) const;

    Signal1<int> sig_ship_changed;
    Signal0      sig_planet_changed;
    Signal0      sig_structure_changed;

    enum {
        MAX_SHIPS = ::NUM_SHIPS
    };
};

/*@}*/


/** Mark this object changed. */
inline void
GSimObject::setChanged()
{
    changed = true;
}

/** Mark this object unchanged. */
inline void
GSimObject::setUnchanged()
{
    changed = false;
}

/** Check whether this object was changed. */
inline bool
GSimObject::isChanged() const
{
    return changed;
}

/** Get object Id. */
inline int
GSimObject::getId() const
{
    return id;
}

/** Change object Id. */
inline void
GSimObject::setId(int id)
{
    this->id = id;
    setChanged();
}

/** Get damage level. */
inline int
GSimObject::getDamage() const
{
    return damage;
}

/** Change damage level. */
inline void
GSimObject::setDamage(int damage)
{
    this->damage = damage;
    setChanged();
}

/** Get shield level. */
inline int
GSimObject::getShield() const
{
    return shield;
}

/** Change shield level. */
inline void
GSimObject::setShield(int shield)
{
    this->shield = shield;
    setChanged();
}

/** Get owner. */
inline int
GSimObject::getOwner() const
{
    return owner;
}

/** Change owner. */
inline void
GSimObject::setOwner(int owner)
{
    this->owner = owner;
    setChanged();
}

/** Get experience level. */
inline int
GSimObject::getExperienceLevel() const
{
    return experience_level;
}

/** Change experience level. */
inline void
GSimObject::setExperienceLevel(int experience_level)
{
    this->experience_level = experience_level;
    setChanged();
}

/** Get flags (fl_XXX). */
inline int32_t
GSimObject::getFlags() const
{
    return flags;
}

/** Change flags (fl_XXX). */
inline void
GSimObject::setFlags(int32_t flags)
{
    this->flags = flags;
    setChanged();
}

/** Get FLAK rating override. */
inline int32
GSimObject::getFlakRatingOverride() const
{
    return flak_r_override;
}

/** Set FLAK rating override. Must be enabled in flags to be actually
    used for simulation. */
inline void
GSimObject::setFlakRatingOverride(int32 r)
{
    flak_r_override = r;
    setChanged();
}

/** Get FLAK compensation override. */
inline int32
GSimObject::getFlakCompensationOverride() const
{
    return flak_c_override;
}

/** Change FLAK compensation override. Must be enabled in flags to be
    actually used for simulation. */
inline void
GSimObject::setFlakCompensationOverride(int32 r)
{
    flak_c_override = r;
    setChanged();
}

/** Check whether this ship has any non-standard function configured. Returns both
    functions explicitly enabled and disabled. */
inline bool
GSimObject::hasAnyNonstandardFunction() const
{
    return (getFlags() & fl_FunctionSetBits) != 0;
}

/** Get crew. */
inline int
GSimShip::getCrew() const
{
    return crew;
}

/** Get hull type. Can be 0 for custom ships. */
inline int
GSimShip::getHull() const
{
    return hull;
}

/** Get mass. Always kept up-to-date. */
inline int
GSimShip::getMass() const
{
    return mass;
}

/** Get beam type. */
inline int
GSimShip::getBeamType() const
{
    return beam_type;
}

/** Get beam count. */
inline int
GSimShip::getBeamCount() const
{
    return beam_count;
}

/** Get torpedo type. */
inline int
GSimShip::getTorpType() const
{
    return torp_type;
}

/** Get torpedo launcher count. */
inline int
GSimShip::getTorpLauncherCount() const
{
    return torp_lcount;
}

/** Get number of fighter bays. */
inline int
GSimShip::getBayCount() const
{
    return bay_count;
}

/** Get ammunition (torps, fighters). */
inline int
GSimShip::getAmmo() const
{
    return ammo;
}

/** Get engine type. */
inline int
GSimShip::getEngineType() const
{
    return engine_type;
}

/** Get Id for intercept-attack. */
inline int
GSimShip::getInterceptId() const
{
    return intercept_id;
}

/** Check whether this is a custom ship. Custom ships have some more
    freedom in configuring. */
inline bool
GSimShip::isCustomShip() const
{
    return getHull() == 0;
}

/** Get number of planetary defense posts. */
inline int
GSimPlanet::getDefense() const
{
    return defense;
}

/** Get number of base defense posts. */
inline int
GSimPlanet::getBaseDefense() const
{
    return base_defense;
}

/** Get starbase beam tech level. Zero means no base. */
inline int
GSimPlanet::getBaseBeamTech() const
{
    return beam_tech;
}

/** Get starbase torpedo tech level. */
inline int
GSimPlanet::getBaseTorpTech() const
{
    return torp_tech;
}

/** Get number of starbase fighters. */
inline int
GSimPlanet::getBaseFighters() const
{
    return base_fighters;
}

#endif
