/**
  *  \file int/if/drawingif.cc
  *  \brief Interpreter Game Interface: Drawings
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/if/drawingif.h"
#include "cpluslib/misc.h"
#include "int/if/ifutil.h"
#include "int/interror.h"
#include "client/marks.h"
#include "int/propenum.h"
#include "game/global.h"
#include "int/args.h"
#include "int/compound.h"
#include "int/exec.h"

class IntExecutionContext;

IntContext*
IFDrawingMake()
{
    if (!haveDisplayedTurn()) {
        return 0;
    }

    GDrawingContainer& drawings = getDisplayedTurn().getCurrentUniverse().getDrawings();
    if (drawings.begin() != drawings.end()) {
        return new IntDrawingContext(drawings);
    } else {
        return 0;
    }
}

/**************************** Drawing Methods ****************************/

/* @q Delete (Drawing Command)
   Delete this drawing.
   @since PCC 1.0.14, PCC2 1.99.20 */
static void
IFDrawingDelete(GDrawingContainer& cont, GDrawingContainer::iterator d, IntArgBlock& args)
{
    args.checkArgumentCount(0);
    if (*d != 0) {
        cont.erase(d);
        getDisplayedTurn().getCurrentUniverse().markChanged();
    }
}

/* @q SetComment s:Str (Drawing Command)
   Set drawing comment.
   @see Comment (Drawing Property)
   @since PCC 1.0.14, PCC2 1.99.20 */
static void
IFDrawingSetComment(GDrawingContainer& /*cont*/, GDrawingContainer::iterator d, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    if (*d != 0) {
        setDrawingProperty(**d, idpComment, args.getNext());
    }
}

/* @q SetColor c:Int (Drawing Command)
   Set drawing color.
   @see Color (Drawing Property)
   @since PCC 1.0.14, PCC2 1.99.20 */
static void
IFDrawingSetColor(GDrawingContainer& /*cont*/, GDrawingContainer::iterator d, IntArgBlock& args)
{
    args.checkArgumentCount(1);
    if (*d != 0) {
        setDrawingProperty(**d, idpColor, args.getNext());
    }
}

namespace {
    class IntDrawingMethod : public IntCallableValue {
     public:
        typedef void (*method_t)(GDrawingContainer&, GDrawingContainer::iterator, IntArgBlock&);

        IntDrawingMethod(GDrawingContainer& cont, GDrawingContainer::iterator d, method_t m);
        ~IntDrawingMethod();

        // IntCallableValue:
        void      call(IntExecutionContext& exc, IntDataSegment& args, bool want_result);
        bool      isProcedureCall();
        int32_t   getDimension(int32_t);
        IntContext* makeFirstContext();
        
        // IntValue:
        string_t    toString(bool readable);
        void        store(IntTagNode&, Stream&);
        IntValue*   clone();

     private:
        GDrawingContainer& cont;
        GDrawingContainer::iterator it;
        method_t m;
    };
}

IntDrawingMethod::IntDrawingMethod(GDrawingContainer& cont, GDrawingContainer::iterator d, method_t m)
    : cont(cont), it(d), m(m)
{ }

IntDrawingMethod::~IntDrawingMethod()
{ }

void
IntDrawingMethod::call(IntExecutionContext& exc, IntDataSegment& args, bool want_result)
{
    IntArgBlock a(args, 0, args.size());
    m(cont, it, a);
    if (want_result) {
        exc.pushNewValue(0);
    }
}

bool
IntDrawingMethod::isProcedureCall()
{
    return true;
}

int32_t
IntDrawingMethod::getDimension(int32_t)
{
    return 0;
}

IntContext*
IntDrawingMethod::makeFirstContext()
{
    throw IntError::typeError(IntError::tyIterable);
}

// IntValue:
string_t
IntDrawingMethod::toString(bool /*readable*/)
{
    return "#<obj-procedure>";
}

void
IntDrawingMethod::store(IntTagNode&, Stream&)
{
    throw IntError::notSerializable();
}

IntValue*
IntDrawingMethod::clone()
{
    return new IntDrawingMethod(*this);
}

/*************************** IntDrawingContext ***************************/

namespace {
    enum { DrawingProperty, DrawingMethod };
}

static const TIntMapping drawing_mapping[] = {
    { "COLOR",        idpColor,      DrawingProperty,    thInt },
    { "COMMENT",      idpComment,    DrawingProperty,    thString },
    { "DELETE",       0,             DrawingMethod,      thProcedure },
    { "END.X",        idpEndX,       DrawingProperty,    thInt },
    { "END.Y",        idpEndY,       DrawingProperty,    thInt },
    { "EXPIRE",       idpExpire,     DrawingProperty,    thInt },
    { "LOC.X",        idpLocX,       DrawingProperty,    thInt },
    { "LOC.Y",        idpLocY,       DrawingProperty,    thInt },
    { "RADIUS",       idpRadius,     DrawingProperty,    thInt },
    { "SETCOLOR",     2,             DrawingMethod,      thProcedure },
    { "SETCOMMENT",   1,             DrawingMethod,      thProcedure },
    { "SHAPE",        idpShape,      DrawingProperty,    thInt },
    { "TAG",          idpTag,        DrawingProperty,    thInt },
    { "TYPE",         idpTypeString, DrawingProperty,    thString },
    { "TYPE$",        idpTypeCode,   DrawingProperty,    thInt },
};

static const IntDrawingMethod::method_t drawing_methods[] = {
    IFDrawingDelete,            // 0
    IFDrawingSetComment,        // 1
    IFDrawingSetColor,          // 2
};

IntDrawingContext::IntDrawingContext(GDrawingContainer& drawings)
    : drawings(drawings),
      current(drawings.begin())
{ }

IntDrawingContext::~IntDrawingContext()
{ }

// IntContext:
int32_t
IntDrawingContext::lookup(const IntNameQuery& name)
{
    return lookupName(name, drawing_mapping, countof(drawing_mapping));
}

void
IntDrawingContext::set(int32_t index, IntValue* value)
{
    if (GDrawing* d = *current) {
        switch (drawing_mapping[index].domain) {
         case DrawingProperty:
            setDrawingProperty(*d, IntDrawingProperty(drawing_mapping[index].index), value);
            getDisplayedTurn().getCurrentUniverse().markChanged();
            break;
         default:
            throw IntError::notAssignable();
        }
    } else {
        throw IntError::notAssignable();
    }
}

IntValue*
IntDrawingContext::get(int32_t index)
{
    if (GDrawing* d = *current) {
        switch (drawing_mapping[index].domain) {
         case DrawingProperty:
            return getDrawingProperty(*d, IntDrawingProperty(drawing_mapping[index].index));
         case DrawingMethod:
            return new IntDrawingMethod(drawings, current, drawing_methods[drawing_mapping[index].index]);
         default:
            return 0;
        }
    } else {
        return 0;
    }
}

bool
IntDrawingContext::next()
{
    if (current == drawings.end()) {
        return false;
    } else {
        GDrawingContainer::iterator i = current;
        ++i;
        if (i == drawings.end()) {
            return false;
        } else {
            current = i;
            return true;
        }
    }
}

IntContext*
IntDrawingContext::clone()
{
    return new IntDrawingContext(*this);
}

GObject*
IntDrawingContext::getObject()
{
    return 0;
}

string_t
IntDrawingContext::toString(bool /*readable*/)
{
    return "#<Marker>";
}

void
IntDrawingContext::store(IntTagNode& /*sv*/, Stream& /*aux*/)
{
    throw IntError::notSerializable();
}

void
IntDrawingContext::enumProperties(IntPropertyAcceptor& acceptor)
{
    acceptor.enumMapping(drawing_mapping, countof(drawing_mapping));
}

IntValue*
getDrawingProperty(const GDrawing& d, IntDrawingProperty idp)
{
    switch (idp) {
     case idpColor:
        /* @q Color:Int (Drawing Property)
           Color of this drawing.
           @assignable
           @see SetColor (Drawing Command), NewLine, NewCircle, NewRectangle, NewMarker */
        return makeIntValue(GDrawing::getUserColorFromColor(d.getColor()));
     case idpComment:
        /* @q Comment:Str (Drawing Property)
           Comment of this drawing.
           Displayed for markers.
           @assignable
           @see SetComment (Drawing Command) */
        return makeStringValue(d.getComment());
     case idpEndX:
        /* @q End.X:Int (Drawing Property)
           X location of endpoint.
           Valid for lines and rectangles, EMPTY for others. */
        if (d.getType() == LineDrawing || d.getType() == RectangleDrawing) {
            return makeIntValue(d.getPos2().x);
        } else {
            return 0;
        }
     case idpEndY:
        /* @q End.Y:Int (Drawing Property)
           Y location of endpoint.
           Valid for lines and rectangles, EMPTY for others. */
        if (d.getType() == LineDrawing || d.getType() == RectangleDrawing) {
            return makeIntValue(d.getPos2().y);
        } else {
            return 0;
        }
     case idpExpire:
        /* @q Expire:Int (Drawing Property)
           Expiration time.
           @assignable
           @see NewLine, NewCircle, NewMarker, NewRectangle */
        return makeIntValue(d.getExpire());
     case idpLocX:
        /* @q Loc.X:Int (Drawing Property)
           X location of starting point/center. */
        return makeIntValue(d.getPos().x);
     case idpLocY:
        /* @q Loc.Y:Int (Drawing Property)
           Y location of starting point/center. */
        return makeIntValue(d.getPos().y);
     case idpRadius:
        /* @q Radius:Int (Drawing Property)
           Radius of drawing.
           Valid for circles, EMPTY for others.
           @assignable */
        if (d.getType() == CircleDrawing) {
            return makeIntValue(d.getCircleRadius());
        } else {
            return 0;
        }
     case idpShape:
        /* @q Shape:Int (Drawing Property)
           Marker shape.
           Valid for markers, EMPTY for others.
           @assignable */
        if (d.getType() == MarkerDrawing) {
            return makeIntValue(d.getMarkerKind());
        } else {
            return 0;
        }
     case idpTag:
        /* @q Tag:Int (Drawing Property)
           Marker tag.
           Usually an integer created with {Atom()}.
           @assignable */
        return makeIntValue(d.getTag());
     case idpTypeString:
        /* @q Type:Str (Drawing Property)
           Type of drawing.
           @see Type$ (Drawing Property) */
        {
            static const char*const markerTypes[] = { "Line", "Rectangle", "Circle", "Marker" };
            uint_t n = d.getType();
            if (n < countof(markerTypes)) {
                return makeStringValue(markerTypes[n]);
            } else {
                return 0;
            }
        }
     case idpTypeCode:
        /* @q Type$:Int (Drawing Property)
           Type of drawing.
           <table align="left">
            <tr><th width="4">Type$</th><th width="5">Type</th></tr>
            <tr><tn>0</tn>              <td>Line</td></tr>
            <tr><tn>1</tn>              <td>Rectangle</td></tr>
            <tr><tn>2</tn>              <td>Circle</td></tr>
            <tr><tn>3</tn>              <td>Marker</td></tr>
           </table> */
        // GDrawingType intentionally matches the chartX.cc type tag,
        // which is what PCC 1.x exposes to the script side, and so do we.
        return makeIntValue(d.getType());
    }
    return 0;
}

void
setDrawingProperty(GDrawing& d, IntDrawingProperty idp, IntValue* value)
{
    int32_t i;
    string_t s;
    switch (idp) {
     case idpColor:
        if (checkIntArg(i, value, 0, GDrawing::NUM_USER_COLORS)) {
            d.setColor(i);
        }
        break;
     case idpComment:
        if (checkStringArg(s, value)) {
            if (d.getType() != MarkerDrawing) {
                throw IntError::notAssignable();
            }
            d.setComment(s);
        }
        break;
     // case idpEndX: - FIXME: could be assignable
     // case idpEndY: - FIXME: could be assignable
     case idpExpire:
        if (checkIntArg(i, value, -1, 0x7FFF)) {
            d.setExpire(i);
        }
        break;
     // case idpLocX: - FIXME: could be assignable 
     // case idpLocY: - FIXME: could be assignable
     case idpRadius:
        if (checkIntArg(i, value, 1, 5000)) {
            if (d.getType() != CircleDrawing) {
                throw IntError::notAssignable();
            }
            d.setCircleRadius(i);
            getDisplayedTurn().getChartBBox().addDrawing(d);
        }
        break;
     case idpShape:
        if (checkIntArg(i, value, 0, NUM_USER_MARKERS-1)) {
            if (d.getType() != MarkerDrawing) {
                throw IntError::notAssignable();
            }
            d.setMarkerKind(i);
        }
        break;
     case idpTag:
        if (checkIntArg(i, value, 0, 0xFFFF)) {
            d.setTag(i);
        }
        break;
     default:
        throw IntError::notAssignable();
    }
}
