/**
  *  \file u/t_game_fcode.cc
  *
  *  (c) 2001-2018 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/fcode.h"
#include "u/t_game.h"
#include "game/host.h"
#include "io/memstream.h"

void
GameFcodeTestSuite::testFCode()
{
    GFCode mkt("mkt", "sc,make torps");
    GFCode lfm("lfm", "sc+9ab,make fighters");
    GFCode att("ATT", "p,attack");

    TS_ASSERT_EQUALS(mkt.getFCode(), "mkt");
    TS_ASSERT(mkt.getRaceMask().contains(1));
    TS_ASSERT(mkt.getRaceMask().contains(2));
    TS_ASSERT(mkt.getRaceMask().contains(10));
    TS_ASSERT_EQUALS(mkt.getDescription(), "make torps");

    TS_ASSERT(!lfm.getRaceMask().contains(1));
    TS_ASSERT(!lfm.getRaceMask().contains(8));
    TS_ASSERT(lfm.getRaceMask().contains(9));
    TS_ASSERT(lfm.getRaceMask().contains(10));
    TS_ASSERT(lfm.getRaceMask().contains(11));

    TS_ASSERT(GFCode::isNumeric("000"));
    TS_ASSERT(!GFCode::isNumeric("00x"));
    TS_ASSERT(!GFCode::isNumeric("x00"));
    TS_ASSERT(!GFCode::isNumeric("+00"));
    TS_ASSERT(GFCode::isNumeric("999"));

    host.clear();
    host.setHostVersion(MKVERSION(3,22,40), ht_Host);
    TS_ASSERT(!GFCode::isNumeric("-11"));
    TS_ASSERT(!GFCode::isNumeric("--1"));
    TS_ASSERT(!GFCode::isNumeric("1"));
    TS_ASSERT(!GFCode::isNumeric(" 1"));
    TS_ASSERT(!GFCode::isNumeric("1 "));
    TS_ASSERT(!GFCode::isNumeric(" 1 "));

    host.setHostVersion(MKVERSION(4,0,0), ht_PHost);
    TS_ASSERT(GFCode::isNumeric("-11"));
    TS_ASSERT_EQUALS(GFCode::getNumericValue("-11"), -11);
    TS_ASSERT(!GFCode::isNumeric("--1"));
    TS_ASSERT(!GFCode::isNumeric("1"));
    TS_ASSERT(!GFCode::isNumeric(" 1"));
    TS_ASSERT(!GFCode::isNumeric("1 "));
    TS_ASSERT(!GFCode::isNumeric(" 1 "));

    host.setHostVersion(MKVERSION(4,0,8), ht_PHost);
    TS_ASSERT(GFCode::isNumeric("-11"));
    TS_ASSERT_EQUALS(GFCode::getNumericValue("-11"), -11);
    TS_ASSERT(!GFCode::isNumeric("--1"));
    TS_ASSERT(GFCode::isNumeric("1"));
    TS_ASSERT_EQUALS(GFCode::getNumericValue("1"), 1);
    TS_ASSERT(GFCode::isNumeric(" 1"));
    TS_ASSERT_EQUALS(GFCode::getNumericValue(" 1"), 1);
    TS_ASSERT(GFCode::isNumeric("1 "));
    TS_ASSERT_EQUALS(GFCode::getNumericValue("1 "), 1);
    TS_ASSERT(GFCode::isNumeric(" 1 "));
    TS_ASSERT_EQUALS(GFCode::getNumericValue(" 1 "), 1);

    host.setHostVersion(MKVERSION(3,4,9), ht_PHost);
    TS_ASSERT(!GFCode::isNumeric(" 1 "));

    host.setHostVersion(MKVERSION(3,4,11), ht_PHost);
    TS_ASSERT(GFCode::isNumeric(" 1 "));
}

void
GameFcodeTestSuite::testRandom()
{
    // references GHost::hasSpacePaddedFCodes(), GHost::hasNegativeFCodes().
    // references GFCode::isExtraFC()
    // references GFCode::isSpecial(), and thus GFCodeList::getInstance().
    // Hence we need some preparations:
    MemoryStream some_extra_fcs(string_t("E zot"));
    some_extra_fcs.seek(0);
    host.setHostVersion(MKVERSION(4, 0, 0), ht_PHost);
    GFCode::clearExtraFC();
    GFCode::loadExtraFC(some_extra_fcs);
    GFCodeList::getInstance().clear();
    GFCodeList::getInstance().addFCode(GFCode("mkt", "sc,make torps"));

    // Now check:
    TS_ASSERT(GFCode::isAllowedRandom("abc"));
    TS_ASSERT(GFCode::isAllowedRandom("01a"));
    TS_ASSERT(GFCode::isAllowedRandom("a01"));
    TS_ASSERT(GFCode::isAllowedRandom("0 1"));  // allowed, but will not be generated
    TS_ASSERT(GFCode::isAllowedRandom("zxy"));
    TS_ASSERT(GFCode::isAllowedRandom("0-1"));  // allowed, but will not be generated
    TS_ASSERT(GFCode::isAllowedRandom("elo"));  // allowed, extra-fc is a PHost thing and thus case-sensitive
    TS_ASSERT(GFCode::isAllowedRandom("Zot"));  // allowed, extra-fc is a PHost thing and thus case-sensitive
    TS_ASSERT(GFCode::isAllowedRandom("zoT"));
    TS_ASSERT(GFCode::isAllowedRandom("zo "));

    TS_ASSERT(!GFCode::isAllowedRandom("mkt")); // case-variants
    TS_ASSERT(!GFCode::isAllowedRandom("Mkt"));
    TS_ASSERT(!GFCode::isAllowedRandom("mKt"));
    TS_ASSERT(!GFCode::isAllowedRandom("mkT"));
    TS_ASSERT(!GFCode::isAllowedRandom("MKT"));
    TS_ASSERT(!GFCode::isAllowedRandom("aab"));
    TS_ASSERT(!GFCode::isAllowedRandom("aba"));
    TS_ASSERT(!GFCode::isAllowedRandom("baa"));
    TS_ASSERT(!GFCode::isAllowedRandom("mf1"));
    TS_ASSERT(!GFCode::isAllowedRandom("mff"));
    TS_ASSERT(!GFCode::isAllowedRandom("MFx"));  // case-insensitive in THost!
    TS_ASSERT(!GFCode::isAllowedRandom("Mfx"));  // case-insensitive in THost!
    TS_ASSERT(!GFCode::isAllowedRandom("mFx"));  // case-insensitive in THost!
    TS_ASSERT(!GFCode::isAllowedRandom("xyz"));
    TS_ASSERT(!GFCode::isAllowedRandom("Xyz"));
    TS_ASSERT(!GFCode::isAllowedRandom("000"));
    TS_ASSERT(!GFCode::isAllowedRandom("012"));
    TS_ASSERT(!GFCode::isAllowedRandom("-19"));
    TS_ASSERT(!GFCode::isAllowedRandom("Elo"));
    TS_ASSERT(!GFCode::isAllowedRandom("Eex"));
    TS_ASSERT(!GFCode::isAllowedRandom("zot"));

    TS_ASSERT(!GFCode::isAllowedRandom("?xy"));
    TS_ASSERT(!GFCode::isAllowedRandom("z?y"));
    TS_ASSERT(!GFCode::isAllowedRandom("zx?"));
    TS_ASSERT(!GFCode::isAllowedRandom("#xy"));
    TS_ASSERT(!GFCode::isAllowedRandom("z#y"));
    TS_ASSERT(!GFCode::isAllowedRandom("zx#"));
    TS_ASSERT(!GFCode::isAllowedRandom("###"));
    TS_ASSERT(!GFCode::isAllowedRandom("?""?""?"));

    TS_ASSERT(!GFCode::isAllowedRandom(""));
    TS_ASSERT(!GFCode::isAllowedRandom("a"));
    TS_ASSERT(!GFCode::isAllowedRandom("ab"));
    TS_ASSERT(!GFCode::isAllowedRandom("abcd"));
}
